package edu.wpi.ari.glados.ins;
import java.util.Vector;

import javax.vecmath.Vector3d;


public class INSManager implements IINSObervable, INSNetObserver {
	
	private INSConnector connection;
	
	private Integrator calcX;
	private Integrator calcY;
	private Integrator calcZ;
	
	private Vector<INSObserver> listeners;
	
	/**
	 * Default constructor
	 * 
	 * @param host the host to connect to
	 * @param port the port to connect on
	 */
	public INSManager(String host, int port) {
		connection = new INSConnector(host, port);
		connection.addListener(this);
		
		calcX = new Integrator();
		calcY = new Integrator();
		calcZ = new Integrator();
		
		listeners = new Vector<INSObserver>();
	}
	
	/**
	 * Set the connection to a new host
	 * 
	 * @param h the new host
	 */
	public void setHost(String h) {
		connection.setHost(h);
	}

	/**
	 * Set the port to connect on
	 * 
	 * @param p the new port
	 */
	public void setPort(int p) {
		connection.setPort(p);
	}
	
	/**
	 * Connect to the INS host
	 */
	public void connect() {
		connection.connect();
	}
	
	/**
	 * Disconnect from the INS host
	 */
	public void disconnect() {
		connection.disconnect();
	}
	
	public void update(double timestep, double Ax, double Ay, double Az, double gyro) {
		//Save the last timestep.
		double last = calcX.getLastX();

		//Calc the new Velocities.
		double VX = calcX.calcV(timestep, Ax);
		double VY = calcY.calcV(timestep, Ay);
		double VZ = calcZ.calcV(timestep, Az);
		
		//Create the result.
		Vector3d result = new Vector3d(VX, VY, VZ);
		
		//FIXME: WE're scaleing up to meters per timestep. SLAM seems to want this,
		//And there's probably a more efficant way to do it (like calculate with time-
		//steps instead of ms in calcV.
		result.scale(calcX.getLastX() - last);
		
		//Get the gyro yaw.
		double theta = Math.toRadians(gyro* 0.9765625); //THETA MUST BE IN RADIANS
		
		//result.x = Math.cos(theta)*result.x - Math.sin(theta)*result.y; 
		//result.y = Math.sin(theta)*result.x + Math.cos(theta)*result.y;
		
		//Fire Listener
		fireListener(result, theta);
	}

	/**
	 * Add a network listener
	 * @param l
	 */
	public void addNetListener(INSNetObserver l) {
		connection.addListener(l);
	}
	
	/**
	 * Remove the network listener
	 * @param l
	 */
	public void removeNetListener(INSNetObserver l) {
		connection.removeListener(l);
	}

	public void addListener(INSObserver l) {
		listeners.add(l);
	}

	public void removeAll() {
		listeners.clear();
	}
	
	public void removeListener(INSObserver l) {
		listeners.remove(l);
	}
	
	private void fireListener(Vector3d v, double yaw) {
		for(INSObserver l : listeners) {
			l.INSReceivedValues(v, yaw);
		}
	}
}
