package edu.wpi.ari.glados.logging;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Vector;

import javax.swing.JCheckBoxMenuItem;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.vecmath.Vector3d;

import edu.wpi.ari.glados.ins.INSObserver;
import edu.wpi.ari.glados.slam.Landmark;
import edu.wpi.ari.glados.slam.PointMap;
import edu.wpi.ari.glados.vision.VisionObserver;

/**
 * Core class for logging implementation
 * Works as a singleton logger that stores all data 
 * used in logging.
 * 
 * Produces a single text log file that can be saved
 * and is actively viewed in a panel
 * @author Alex Cunningham
 *
 */
public class Logger extends JFrame implements INSObserver,
											  VisionObserver,
											  ActionListener,
											  ItemListener{
	
	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;

	/**
	 * Initial time for timestamps
	 */
	private long start_time = 0;
	
	//configuration
	private boolean enableINSLogging;
	private boolean enableVisionLogging;
	
	//GUI Components
	private JTextArea console;
	private JMenuItem saveButton;
	private JCheckBoxMenuItem enLogINS;
	private JCheckBoxMenuItem enLogVis;
	
	
	
	
	private static Logger log = null;
	
	private Logger(){
		super("Logger");
		//build up GUI component
		this.setSize(500, 500);
		
		//configure console
		console = new JTextArea(5, 20);
		JScrollPane scrollpane = new JScrollPane(console);
		this.add(scrollpane);
		
		//Setup menu
		JMenuBar menuBar = new JMenuBar();
		JMenu logMenu = new JMenu("Config");
		menuBar.add(logMenu);
		
		this.enLogVis = new JCheckBoxMenuItem("Enable Vision Logging");
		this.enLogVis.addItemListener(this);
		logMenu.add(this.enLogVis);
		
		this.enLogINS = new JCheckBoxMenuItem("Enable INS Logging");
		this.enLogINS.addItemListener(this);
		logMenu.add(this.enLogINS);
		
		this.saveButton = new JMenuItem("Save Log");
		this.saveButton.addActionListener(this);
		logMenu.add(this.saveButton);
		
		this.add(BorderLayout.NORTH, menuBar);
	
		//initialize logger components
		this.start_time = System.currentTimeMillis();
		
		//Configure cleanup
		this.addWindowListener(new WindowAdapter() 
		{public void windowClosing(WindowEvent e) 
		{System.exit(0);}});
		
		//Turn on GUI
		this.setVisible(true);
	}
	
	/**
	 * Basic Singleton class implementation
	 * @return
	 */
	public static Logger getLogger(){
		if (Logger.log == null)
			return new Logger();
		else 
			return Logger.log;
	}
	/**
	 * Adds a string to the rest of the log.  
	 * Automatically appends the linebreak at the end
	 * This will force a message to appear
	 * @param s
	 */
	public void logString(String s){
		this.console.append("(" + getTimestamp() + ") " + s + "\n");
	}
	
	/**
	 * Logs values from INS
	 */
	public void INSReceivedValues(Vector3d step, double yaw) {
		if (this.enableINSLogging)
			this.console.append("(" + getTimestamp() + ") INS: " + step.toString() + "Yaw: " + yaw + "\n");
	}

	/**
	 * Logs values from Vision
	 */
	public void VisionReceivedValues(Vector<Vector3d> points) {
		if (this.enableVisionLogging){
			PointMap map = new PointMap();
			for (int i = 0; i < points.size(); i++){
				map.addLandmark(new Landmark(points.get(i)));
			}
			this.console.append("(" + getTimestamp() + ") Vision: " + map.toString() + "\n");	
		}
	}
	
	private long getTimestamp(){
		return System.currentTimeMillis() - this.start_time;
	}
	
	/**
	 * Handler for menu buttons
	 */
	public void actionPerformed(ActionEvent e) {
		if (e.getSource() == this.saveButton){
			JFileChooser fc = new JFileChooser();
			int ret = fc.showSaveDialog(this);
			if (ret == JFileChooser.APPROVE_OPTION){
				File file = fc.getSelectedFile();
				PrintWriter out = null;
				
				try {
					out = new PrintWriter(new FileWriter(file));
				} catch (IOException e1) {
					e1.printStackTrace();
				}
				out.print(this.console.getText());
				out.close();
			}
		}
		
	}

	/**
	 * Handler for check boxes in the menu
	 */
	public void itemStateChanged(ItemEvent e) {
		if (e.getSource() == this.enLogINS){
			this.enableINSLogging = this.enLogINS.getState();
		} else if (e.getSource() == this.enLogVis) {
			this.enableVisionLogging = this.enLogVis.getState();
		}
		
	}
	
	private String getConsoleText(){
		return this.console.getText();
	}
	
	/**
	 * This is for testing only!!!!!
	 * @param args
	 */
	public static void main(String [] args){
		Logger log = Logger.getLogger();
		log.logString("This is a Test!");
		log.logString("Another line!");
		System.out.println(log.getConsoleText());
		
	}

}
