package edu.wpi.ari.glados.slam;


import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.util.Vector;

import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.vecmath.Vector3d;

import edu.wpi.ari.glados.ins.INSObserver;
import edu.wpi.ari.glados.slam.particle.Particle;
import edu.wpi.ari.glados.slam.slamIDE.ConsoleOutput;
import edu.wpi.ari.glados.slam.slamIDE.FixCoords;
import edu.wpi.ari.glados.slam.slamIDE.ScaleFilter;
import edu.wpi.ari.glados.vision.VisionObserver;

/**
 * Wrapper GUI element for the SLAM visualization
 * Contains a full SLAM implementation, and needs
 * to have data from the rest of the system
 * 
 * Data assumed to come in at realtime speeds,
 * and that timestamps can be taken when functions
 * are called.  
 * @author Alex Cunningham
 *
 */
public class SLAMFrame extends JFrame implements INSObserver, 
												 VisionObserver,
												 ConsoleOutput{
	
	private static final long serialVersionUID = 1L;
	
	/*************************** Data ************************/

	/**
	 * Basic SLAM component - contains basic processing functionality
	 */
	private Particle particle;
	
	/**
	 * Text console
	 */
	private JTextArea console;
	
	/**
	 * Visualization Panel
	 */
	private SLAMVisPanel vis;

	/**
	 * File for logging purposes
	 */
	private SLAMfile datalog;

	

	/*************************** Constructors ************************/
	
	/**
	 * Basic constructor, loads with defaults
	 */
	public SLAMFrame(){
		//initialize SLAM engine
		this.particle = new Particle();
		//create GUI components
		
		//setup file
		this.datalog = new SLAMfile();
		
		//configure console
		console = new JTextArea(5, 20);
		JScrollPane scrollpane = new JScrollPane(console);
		this.add(BorderLayout.SOUTH, scrollpane);
		
		//setup VisPanel
		this.vis = new SLAMVisPanel(this.particle);
		JScrollPane map_pane = new JScrollPane(vis);
		this.add(BorderLayout.CENTER, map_pane);
		vis.init();
		
		//Show basic initialization message
		this.console.append("SLAM Initialization Complete!\n");
		this.console.setEditable(false);
		this.setSize(500, 500);
		
		//set visible
		this.setVisible(true);
	}
	
	/*************************** Listener Handlers ************************/
	
	/**
	 * Handles calls by INS Manager
	 * @param step 3D translational change since last timestep in meters
	 * @param yaw rotation around z axis since last time step in radians
	 */
	public void INSReceivedValues(Vector3d step, double yaw) {
		//System.out.println("In INSReceived Values: " + step.toString());
		Control ctrl = new Control(step.x, step.y, step.z, 0, 0, yaw);
		updateControl(ctrl);
	}
	
	/**
	 * Handles calls by Vision Manager
	 * @param points set of visible points determined by vision
	 */
	public void VisionReceivedValues(Vector<Vector3d> points) {
		//System.out.println("In Vision Received: " + points.toString());
		PointMap map = new PointMap();
		for (int i = 0; i < points.size(); i++){
			map.addLandmark(new Landmark(points.get(i)));
		}
		updateVision(map);
		updateView();
	}
	
	/*************************** Updating methods ************************/
	
	/**
	 * Refreshes everything with no changes
	 */
	public void updateView(){
		//System.out.println("Updating View: ");
		//get a report from the particle
		String p_report = this.particle.toString();
		this.printToConsole(p_report);
		vis.paint(vis.getGraphics());
	}
	
	/**
	 * Adds a new control to the particle for processing
	 * This will not do a full SLAM processing run, just 
	 * update the current predicted pose
	 * @param ctrl
	 */
	public void updateControl(Control ctrl){
		//System.out.println("From inside updateControl " + ctrl.toString());
		//process particle
		this.datalog.addStim(ctrl);
		this.particle.addControl(ctrl);
	}
	
	/**
	 * Adds a new set of local landmarks to the particle
	 * This will actually start full SLAM processing
	 * @param map
	 */
	public void updateVision(PointMap map){
		//System.out.println("UpdateVision: " + map.toString());
		this.datalog.addStim(map);
		this.particle.addVision(map);		
	}
	
	public void clearDataLog() {
		this.datalog.clear();
		printToConsole("Input log cleared");
	}
	
	public void saveDataLog() {
		JFileChooser fc = new JFileChooser();
		int ret = fc.showSaveDialog(this);
		if (ret == JFileChooser.APPROVE_OPTION){
			File file = fc.getSelectedFile();	
			SLAMfile.saveLog(this.datalog, file, this);
			printToConsole("Input log saved to file: " + file.getName());
		}
	}

	/*************************** Helper Functions ************************/
	
	public void printToConsole(String str){
		this.console.append(str + "\n");
	}
	
	public void replaceConsoleText(String str){
		this.console.setText(str + "\n");
	}
	
	
	/*************************** Testing ************************/
	
	/**
	 *  Basic Testing function
	 * @param args
	 */
	public static void main(String[] args){

		SLAMFrame slam = new SLAMFrame();
		slam.setVisible(true);
		slam.setSize(800, 800);
		slam.setTitle("SLAM Test");
//		fr.setContentPane(slam);
		
		//Configure cleanup
		slam.addWindowListener(new WindowAdapter() 
		{public void windowClosing(WindowEvent e) 
		{System.exit(0);}});
		
//		fr.setVisible(true);
				
		//add test data
		slam.updateControl(new Control(.2, .2, .2, 0, 0, 0));
		slam.updateControl(new Control(.2, .2, .2, 0, 0, 0));
		PointMap map = new PointMap();
		map.addLandmark(new Landmark(1, 1, 1));
		slam.updateVision(map);
		slam.updateView();
		
		//test file saving

	}

	public void fixCoordDialog(){
		FixCoords fix = FixCoords.getFixCoords();
		fix.showFixCoordsDialog();
	}
	
	public void setScaleDialog(){
		ScaleFilter sf = ScaleFilter.getScaleFilter();
			sf.showDialog();
	}
	
}
