package edu.wpi.ari.glados.slam;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
//import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Line2D;
import java.awt.geom.Rectangle2D;
import java.util.Vector;

import javax.swing.JFrame;
import javax.swing.JPanel;

import edu.wpi.ari.glados.slam.particle.Particle;


/**
 * Rendering panel for SLAM Module
 * 
 * Major Update: Adding pan controls
 * @author Alex Cunningham
 *
 */
public class SLAMVisPanel extends JPanel implements MouseMotionListener, MouseListener{
	
	private static final long serialVersionUID = 1L;
	
	final static Color bg = Color.white;
    final static Color fg = Color.black;
    final static BasicStroke stroke = new BasicStroke(2.0f);
    
    /**
     * Link to particle, used for pulling data out
     */
    private Particle particle;
    
    /**
     * Color Scheme 
     */
    private SLAMVisColorScheme scheme;
    
    //Mouse Coords
    private double x_pos = 0;
    private double y_pos = 0;
    
    //Record View Settings
    /**
     * Number of pixels per meter
     * Functions as the zoom for the system view, with small numbers
     * representing zoomed out conditions
     * Base zoom is always used, with an additional
     * multiplier to handle zoom factors
     */
    final double base_zoom = 100.; 
    private double scale = 2.;
    
    //Track center of viewable area in system coordinates
    private double x_center;
    private double y_center; 
    
    /**
     * Not to be used
     */
    //private SLAMVisPanel(){}
    
    /**
     * Primary Constructor
     * - needs to have a particle in order to have
     * information to render
     * @param p
     */
    public SLAMVisPanel(Particle p){
    	this.addMouseMotionListener(this);
    	this.addMouseListener(this);
    	this.particle = p;
    	this.setOpaque(true);
    }
	
	public SLAMVisPanel() {
		this.addMouseMotionListener(this);
    	this.addMouseListener(this);
    	this.particle = new Particle();
    	this.setOpaque(true);
	}

	public void init(){
		//Initialize drawing colors
        setBackground(bg);
        setForeground(fg);
        this.setSize(800, 800);
        this.scheme = SLAMVisColorScheme.getPrintDefault();
        
        this.x_center = 0.;
        this.y_center = 0.;
	}
	
	public void paint(Graphics g){
		Graphics2D g2 = (Graphics2D) g;
		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		
		//get size for reference
		Dimension d = this.getSize();
		int height = d.height;
		int width = d.width;
		
		//render dark background
		g2.setPaint(this.scheme.getBkgnd());
		Rectangle2D.Double background = new Rectangle2D.Double(0., 0., (double)width, (double) height);
		g2.draw(background);
		g2.fill(background);
		
		//render axes
		g2.setPaint(this.scheme.getAxis_lines());
		Line2D.Double x_axis, y_axis;
		double y_cross = (double)transformY(0);
		double x_cross = (double)transformX(0);
		if (y_cross > -1) {
			x_axis = new Line2D.Double(0., y_cross, (double) width, y_cross);
			g2.draw(x_axis);
		}
		if (x_cross > -1) {
			y_axis = new Line2D.Double(x_cross, 0., x_cross, (double) height);
			g2.draw(y_axis);
		}
		//render prior path
		g2.setPaint(this.scheme.getPast_poses());
		for (int i = 0; i < this.particle.getPath().size(); i++){
			Ellipse2D p = drawPose(this.particle.getPath().get(i));
			if (p != null) {
				g2.draw(p);
			}
		}
		
		//render current pose
		g2.setPaint(this.scheme.getCur_pose());
		Ellipse2D p = drawPose(this.particle.getPose());
		if (p != null) {
			g2.draw(p);
		}
		
		//render landmarks in map
		g2.setPaint(this.scheme.getLandmarks());
		for (int i = 0; i < this.particle.getMap().getNumLandmarks(); i++){
			Ellipse2D lm = drawLandmark(this.particle.getMap().getLandmark(i));
			if (lm != null) {
				g2.fill(lm);
			}
		}
		
		//add labels
		g2.setPaint(this.scheme.getText());
		// *** THIS WAS CHANGED... MAKE IT WORK LATER ***
		//Font font = new Font(Font.DIALOG, Font.PLAIN, 12);
		//g2.setFont(font);
		g2.drawString("X: " + this.x_pos + " Y: " + this.y_pos, 0, height);
		g2.drawString("Center: (" + this.x_center + ", " + this.y_center + ")", width/2, height);

	}
	
	//Event Handlers
	
	/**
	 * Tracks the position of the mouse
	 */
	public void mouseMoved(MouseEvent e){
		//update cursor position
		double new_x = pixel2globalX(e.getX());
		double new_y = pixel2globalY(e.getY());
		
		//Check for a change
		if (new_x != this.x_pos || new_y != this.y_pos){
			this.x_pos = new_x;
			this.y_pos = new_y;
			this.drawText();
		}
	}
	
	/**
	 * Performs panning operation - does nothing
	 */
	public void mouseDragged(MouseEvent e){
		
	}
	
	/**
	 * Centers the screen on the location of the point
	 * Popup trigger toggles color scheme
	 * @param e
	 */
	public void mouseClicked(MouseEvent e) {
		if (e.getButton()==MouseEvent.BUTTON1){
			this.x_center = this.pixel2globalX(e.getX());
			this.y_center = this.pixel2globalY(e.getY());
			this.paint(this.getGraphics());
		} else if (e.getButton()==MouseEvent.BUTTON3){
			if (this.scheme.equals(SLAMVisColorScheme.getPrintDefault())){
				this.scheme = SLAMVisColorScheme.getScreenDefault();
			} else {
				this.scheme = SLAMVisColorScheme.getPrintDefault();
			}
			this.paint(this.getGraphics());
		}
	}


	public void mouseEntered(MouseEvent e) {
		// TODO Auto-generated method stub
		
	}

	public void mouseExited(MouseEvent e) {
		// TODO Auto-generated method stub
		
	}

	public void mousePressed(MouseEvent e) {
		//System.out.println("Mouse Pressed...");
		
	}

	public void mouseReleased(MouseEvent e) {
		//System.out.println("... Mouse Released!");
		
	}
	
	
	/**
	 * Takes a pose and generates the visual representation 
	 * @param pose
	 * @param p_ycenter 
	 * @param p_xcenter 
	 * @return
	 */
	private Ellipse2D drawPose(Pose pose){
		//pose dot parameters
		final double r = 10;
		
		int x = this.transformX(pose.getX());
		int y = this.transformY(pose.getY());
		
		if (x != -1 && y != -1)
			return new Ellipse2D.Double(x, y, r, r);
		else 
			return null;
	}
	
	/**
	 * Takes a Landmark and generates a rendered view
	 * @param lm
	 * @param p_ycenter 
	 * @param p_xcenter 
	 * @return
	 */
	private Ellipse2D drawLandmark(Landmark lm){
		//pose dot parameters
		final double r = 5;
		
		int x = this.transformX(lm.getX());
		int y = this.transformY(lm.getY());
		
		if (x != -1 && y != -1)
			return new Ellipse2D.Double(x, y, r, r);
		else 
			return null;
	}
	
	private void drawText(){
		Graphics2D g2 = (Graphics2D) this.getGraphics();
		int height = this.getHeight();
		int width = this.getWidth();
		//draw background box
		g2.setPaint(this.scheme.getBkgnd());
		Rectangle2D.Double matte = new Rectangle2D.Double(0., (double)(height-13), (double)(width/2), 13);
		g2.fill(matte);
		
		//draw text
		g2.setPaint(this.scheme.getText());
		g2.drawString("X: " + this.x_pos + " Y: " + this.y_pos, 0, height);
	}
	
	/**
	 * Scales an x coordinate to get pixel size
	 * @param x global x coordinate to scale
	 * @return a pixel position using the zoom factors
	 */
	public int scaleX(double x){
		return (int)(x*this.base_zoom*this.scale);
	}
	
	/**
	 * Scales a y coordinate to get pixel size.
	 * Inverts to handle change system change
	 * @param y global y coordinate to scale
	 * @return a pixel position using zoom factors
	 */
	public int scaleY(double y){
		return -1*(int)(y*this.base_zoom*this.scale);
	}
	
	/**
	 * Converts a pixel position into a global coordinate
	 * position
	 * @param x
	 * @return
	 */
	public double pixel2globalX(int x){
		//remove width adjustment
		int offset = x - this.getWidth()/2;
		//add center
		int center = scaleX(this.x_center);
		offset += center;
		//remove scale
		return ((double)offset)/(this.base_zoom*this.scale);
	}
	
	/**
	 * Converts a pixel position into a global coordinate
	 * @param y
	 * @return
	 */
	public double pixel2globalY(int y){
		//remove width adjustment
		int offset = y - this.getHeight()/2;
		//add center
		int center = scaleY(this.y_center);
		offset += center;
		//remove scale
		return -1*((double)offset)/(this.base_zoom*this.scale);
	}
	
	/**
	 * Converts an x coordinate from a global (double) representation 
	 * to a pixel location that can be rendered.  This needs to be 
	 * used immediately in the redraw process as this method is 
	 * dependent on the center, scaling factors and size of panel
	 * 
	 * NOTE: returns -1 if out of range
	 * @param global X coordinate in real space
	 * @return x coordinate in pixels ready for mapping
	 */
	public int transformX(double global){
		//convert global to pixels using zoom parameter
		int x  = scaleX(global);
		//convert center to pixels using zoom parameter
		int c = scaleX(this.x_center);
		//use dimensions of screen to adjust center
		int w = this.getWidth();
		//determine if global point is viewable
		int offset = x-c;
		if (Math.abs(offset)<(w/2)){
			return w/2 + offset;
		} else 
			return -1;
	}
	
	/**
	 * Converts an y coordinate from a global (double) representation 
	 * to a pixel location that can be rendered.  This needs to be 
	 * used immediately in the redraw process as this method is 
	 * dependent on the center, scaling factors and size of panel
	 * 
	 * NOTE: returns -1 if out of range
	 * @param global Y coordinate in real space
	 * @return y coordinate in pixels ready for mapping
	 */
	public int transformY(double global){
		//convert global to pixels using zoom parameter
		int y  = scaleY(global);
		//convert center to pixels using zoom parameter
		int c = scaleY(this.y_center);
		//use dimensions of screen to adjust center
		int h = this.getHeight();
		//determine if global point is viewable
		int offset = y-c;
		if (Math.abs(offset)<(h/2)){
			return h/2 + offset;
		} else 
			return -1;
	}
	
	public void setCenter(double x, double y){
		this.x_center = x;
		this.y_center = y;
	}
	
	/**
	 * Test version of 
	 * @param args
	 */
	public static void main(String [] args){
		//setup particle
		Particle p = new Particle();
		//fill map
		PointMap map = new PointMap();
		for (int i = 0; i < 20; i++){
			map.addLandmark(Landmark.randLandmark(10., 5.));
		}
		//fill path
		Vector<Pose> path = new Vector<Pose>();
		for (int i = 0; i < 10; i++){
			path.add(new Pose(i*.5, i*.5, 0, 0, 0, 0));
		}
		//set current pose
		p.setPath(path);
		p.setPose(new Pose(0, 0, 0, 0, 0, 0));
		p.setMap(map);
		
		//setup vis panel 
		SLAMVisPanel panel = new SLAMVisPanel(p);
		panel.init();
		panel.setCenter(1.0, 0);
		
		JFrame fr = new JFrame();
		fr.add(panel);
		fr.setSize(800, 800);
		fr.setVisible(true);
		
		//Configure cleanup
		fr.addWindowListener(new WindowAdapter() 
		{public void windowClosing(WindowEvent e) 
		{System.exit(0);}});
		
	}


	

}
