package edu.wpi.ari.glados.slam.particle;

import java.util.Vector;
import javax.vecmath.Point3d;
import javax.media.j3d.Transform3D;

import edu.wpi.ari.glados.slam.Control;
import edu.wpi.ari.glados.slam.Landmark;
import edu.wpi.ari.glados.slam.PointMap;
import edu.wpi.ari.glados.slam.Pose;
import edu.wpi.ari.glados.slam.Stim;
import edu.wpi.ari.glados.slam.slamIDE.ConsoleOutput;
import edu.wpi.ari.glados.slam.slamIDE.FixCoords;
import edu.wpi.ari.glados.slam.slamIDE.ScaleFilter;

/**
 * 
 * @author Alex Cunningham
 * Basic element of a particle filter, stores the map and robot pose, as
 * well as uncertainty measurements and other related tag information
 */
public class Particle {
	
	//-----------------------------------------------------------------------
	//
	// Member Fields
	
	/**
	 * Current pose of robot
	 */
	protected Pose pose;
	/**
	 * Map to generate
	 */
	protected PointMap map;
	/**
	 * Path of robot prior to the current time step
	 */
	protected Vector<Pose> path;
	
	/**
	 * Control to be added over time while waiting for vision input
	 */
	private Control tot_ctrl;
	
	/**
	 * Coordinate fixing filter
	 */
	private FixCoords fixcoords;
	
	/**
	 * Scaling filter
	 */
	private ScaleFilter scale;
	
	/**
	 * Link to a block with a console output
	 */
	private ConsoleOutput out;
	
	//*************** Parameters for the filter
	private double DA_THRESH;
	private final double VISION_AVU_TO_METERS = .001; 
	private final double INS_AVU_TO_METERS = .01;
	//processing type
	public enum proc_type {BASIC, KALMAN}
	private proc_type cur_proc;
	//data association type
	public enum da_type {SIMPLE, ML}
	private da_type cur_da;
	
	
	
	
	
	
	//-----------------------------------------------------------------------
	//
	// Constructors
	
	/**
	 * Configures the particle with defaults:
	 * DA_THRESH = 0.5m
	 * DA Type = SIMPLE
	 * Processing Type = BASIC
	 */
	public Particle(){
		this.pose = new Pose (0, 0, 0, 0, 0, 0);
		this.map = new PointMap();
		this.path = new Vector<Pose>();
		this.tot_ctrl = new Control();
		this.DA_THRESH = .5;
		this.cur_da = Particle.da_type.SIMPLE;
		this.cur_proc = Particle.proc_type.BASIC;
		this.fixcoords = FixCoords.getFixCoords();
		this.scale = ScaleFilter.getScaleFilter();
	}

	//-----------------------------------------------------------------------
	//
	// Processing Methods
	
	
	/**
	 * Adds a generic stim to the particle, with 
	 * full processing only if necessary.  Also performs 
	 * coordinate switching and scaling
	 * @param s - a generic Stim that needs to be 
	 * identified and processing occurs if it is possible
	 */
	public void processStim(Stim s) {
		if (s instanceof Control){
			Control c = (Control)s;
			this.scale.filter(c);
			this.addControl(c);
		} else if (s instanceof PointMap){
			PointMap map = (PointMap)s;
			this.scale.filter(map);
			this.fixcoords.filter(map);
			this.addVision(map);
		}
	}
	
	
	/**
	 * Adds a new control to the particle
	 * Note: this does not execute any real processing, only 
	 * adds the odometry data to the running tally
	 * @param ctrl Control to add to the particle
	 */
	public void addControl(Control ctrl) {
		//this.tot_ctrl.addControl(ctrl);
		// TEMP HACK
		this.tot_ctrl = ctrl;
	}

	/**
	 * Adds a new vision set to the particle, and performs
	 * full SLAM algorithm.  
	 * @param map
	 */
	public void addVision(PointMap map) {
		//process particle
		this.process_basic(this.tot_ctrl, map);
		//reset control
		this.tot_ctrl = new Control();
	}
	
	/**
	 * Switches between processing implementations, this is 
	 * simply a wrapper for other processing functions, based on the
	 * parameter set process type
	 * @param ctrl
	 * @param map
	 */
	public void process(Control ctrl, PointMap map){
		switch(this.cur_proc){
		case BASIC:
			process_basic(ctrl, map);
			break;
		case KALMAN:
			process_kalman(ctrl, map);
			break;
		default:
			this.out.printToConsole("Invalid Processing Method");
		}
	}
	
	protected void process_basic(Control ctrl, PointMap map) {
		//this.out.printToConsole("In Process: " + map);
		
		//store path
		this.path.add(Pose.clone(this.pose));
		
		//update pose
		this.pose.addControl(ctrl);
		
		//scale pose to real-world meters
		this.pose.scale(this.INS_AVU_TO_METERS);
		
		//scale landmarks to real-world meters
		map.scale(this.VISION_AVU_TO_METERS);
		
		//Find global versions of landmarks
		PointMap globals = new PointMap();
		for (int i = 0; i < map.getNumLandmarks(); i++){
			globals.addLandmark(Particle.local2Global(this.pose, map.getLandmark(i)));
		}
		
		//Perform data association
		//TESTING HACK: 
		//PointMap global_unknown = data_associate(globals);
		
		//Update existing map
		//TESTING HACK:
		this.map.addMap(globals);
		//this.map.addMap(global_unknown);
		this.out.printToConsole("Map After Adding new points: " + this.map.toString());
	}
	
	protected void process_kalman(Control ctrl, PointMap map) {
		this.out.printToConsole("Kalman Particle Processing not implemented!");
		//create predicted pose using control
		//Transform measurements into globals
		//perform data association
		//Use ML data association algorithm on existing points
		//loop over known points and check if new measurements are correct
		//sort by distance to robot (closest first)
		//Create a set of possible poses for matching landmarks
		//Perform kalman updates on the predicted pose
		//Set updated pose as the real pose for the robot
		//Recalculate global landmarks from stims
		//perform data association
		//Loop over visible known landmarks
		//perform kalman updates
		//increment score
		//add visible unknown landmarks to map
		//Set particle map
	}
	
	//-----------------------------------------------------------------------
	//
	// Data Association Methods
	
	protected PointMap data_associate(PointMap globals){
		switch(this.cur_da){
		case SIMPLE:
			return data_associate_simple(globals);
		case ML:
			return data_associate_ML(globals);
		default:
			this.out.printToConsole("Invalid Data Association Method");
			return new PointMap();
		}
	}
	
	
	private PointMap data_associate_ML(PointMap globals) {
		this.out.printToConsole("Kalman Particle Processing not implemented!");
		return new PointMap();
	}

	public PointMap data_associate_simple(PointMap globals) {
		//Perform data association
		PointMap global_known = new PointMap();
		PointMap global_unknown = new PointMap();
		for (int i = 0; i < map.getNumLandmarks(); ++i){
			boolean known = false;
			for (int j = 0; j < this.getMap().getLandmarks().size(); j++){
				//check distance threshold with existing landmarks
				//if measurements are close to existing ones
				//add to known and increment landmark scores
				double dist = globals.getLandmarks().get(i).getMean().distance(this.getMap().getLandmarks().get(j).getMean());
				if (dist <= DA_THRESH){
					known = true;
					this.getMap().getLandmark(j).incScore();
				} 
			}
			if (known)
				global_known.addLandmark(globals.getLandmark(i));
			else 
				global_unknown.addLandmark(globals.getLandmark(i));
		}
		return global_unknown;
	}

	//-----------------------------------------------------------------------
	//
	// Static Transforms
	


	/**
	 * Takes a given position and a local landmark and transforms
	 * the local landmark into a global landmark
	 * @param pose
	 * @param local 
	 * @return global landmark
	 */
	public static Landmark local2Global(Pose pose, Landmark local){
		Point3d global = (Point3d) local.getMean().clone();
		Transform3D trans = new Transform3D();
		//add rotation
		//Negative angles to invert them
		trans.rotX(-pose.getXRot());
		trans.transform(global);
		trans.rotY(-pose.getYRot());
		trans.transform(global);
		trans.rotZ(-pose.getZRot());
		trans.transform(global);
		//translational transform
		//Vector3d vec = new Vector3d(pose.getX(), pose.getY(), pose.getZ());
		//this.out.printToConsole("Translation: " + vec.toString());
		//trans.setTranslation(vec);
		//trans.transform(global);
		global.add(pose.getLinPose());
				
		return new Landmark(global);
	}
	
	/**
	 * Calculates a pose based on a local and a global landmark, used for
	 * backsolving the pose of the robot from known landmark positions.
	 * 
	 * Global position comes from subtracting the local vector from the 
	 * global landmark vector.  
	 * 
	 * Angles cannot be calculated from this transform
	 * @param global - global landmark
	 * @param local - local landmark
	 * @return 
	 */
	public static Pose lms2Pose(Landmark global, Landmark local){
		Pose p = new Pose();
		//subtract to get the linear position
		Point3d global_p = new Point3d(global.getMean());
		Point3d local_p = new Point3d(local.getMean());
		global_p.sub(local_p);
		p.setLinPose(new Point3d(global_p));
		//calculate rotations
		return p;
	}
	
	/**
	 * Calculates the local position of a landmark based on a pose and 
	 * a global landmark
	 * Primarily used for generating data
	 * @param pose
	 * @param global
	 * @return
	 */
	public static Landmark global2local(Pose pose, Landmark global){
		Point3d local = (Point3d) global.getMean().clone();
		Transform3D trans = new Transform3D();
		//subtract the pose from the global to get lin position
		local.sub(pose.getLinPose());
		//rotate in positive direction by amount of rotation in pose
		trans.rotX(-pose.getXRot());
		trans.transform(local);
		trans.rotY(-pose.getYRot());
		trans.transform(local);
		trans.rotZ(-pose.getZRot());
		trans.transform(local);
		return new Landmark(local);
	}
	
	
	//-----------------------------------------------------------------------
	//
	// Basic Access Commands
	
	public String toString(){
		return "Pose: " + this.pose.toString() +
		"\nMap: " + this.map.toString();
	}
	
	public void setPose(Pose pose){
		this.pose = pose;
	}
	
	public Pose getPose() {
		return pose;
	}

	public void setMap(PointMap map){
		this.map = map;
	}
	
	public PointMap getMap() {
		return this.map;
	}

	public Vector<Pose> getPath() {
		return path;
	}

	public void setPath(Vector<Pose> path) {
		this.path = path;
	}

	public void clear() {
		this.map = new PointMap();
		this.pose = new Pose();
		this.path.clear();
		this.tot_ctrl = new Control();
		
	}
	
	public void setConsoleOutput(ConsoleOutput out){
		this.out = out;
	}





	
	

	



}
