package edu.wpi.ari.glados.slam.slamIDE;

import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.OutputStreamWriter;
import java.util.Vector;

import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.JTabbedPane;
import javax.swing.JTextArea;

import edu.wpi.ari.glados.slam.Control;
import edu.wpi.ari.glados.slam.PointMap;
import edu.wpi.ari.glados.slam.SLAMVisPanel;
import edu.wpi.ari.glados.slam.Stim;
import edu.wpi.ari.glados.slam.particle.Particle;

/**
 * This is the main app for offline playback of slam
 * datalogs.  It includes a standard visualization 
 * panel, a console log, and a configuration panel
 * for various parts of the system.  
 * @author Alex Cunningham
 *
 */
public class IDEFrame extends JFrame implements ActionListener, ConsoleOutput{

	private static final long serialVersionUID = 1L;

	
	// GUI Components
	private SLAMVisPanel slamvis;
	private JTextArea console;
	private PlaybackControls controls;
	private ParticleConfig config;
	
	// GUI Containers
	private JSplitPane leftPane;
	private JSplitPane mainSplit;
	private JTabbedPane tabs;
	
	// Menu Components
	//   File Menu
	private JMenuItem openDatalog;
	private JMenuItem exportDatalog;
	private JMenuItem exit;
	
	//General FileChooser
	private JFileChooser fc;

	//Datalog
	private Vector<Stim> curLog;
	private String logName;
	
	//SLAM Components
	private Particle particle;
	
	 
	
	
	public IDEFrame(){
		
		//Initialize
		this.curLog = new Vector<Stim>();
		this.logName = "";
		this.fc = new JFileChooser();
		
		//initialize this frame
		this.setSize(800, 800);
		this.setTitle("SLAM Offline Analysis");
		
		//Build Menu
		JMenuBar menubar = new JMenuBar();
		JMenu fileMenu = new JMenu("File");
		
		this.openDatalog = new JMenuItem("Open Datalog");
		fileMenu.add(this.openDatalog);
		this.openDatalog.addActionListener(this);
		
		this.exportDatalog = new JMenuItem("Export Datalog to Text");
		fileMenu.add(this.exportDatalog);
		this.exportDatalog.addActionListener(this);
		
		this.exit = new JMenuItem("Exit");
		fileMenu.add(this.exit);
		this.exit.addActionListener(this);
		
		menubar.add(fileMenu);
		this.setJMenuBar(menubar);
		
		//Initialize the panels
		this.particle = new Particle();
		this.particle.setConsoleOutput(this);
		this.slamvis = new SLAMVisPanel(this.particle);
		this.slamvis.init();
		this.slamvis.setMinimumSize(new Dimension(500,500));
		
		this.console = new JTextArea(5, 20);
		this.console.setEditable(false);
		this.console.setWrapStyleWord(true);
		JScrollPane consoleScroll = new JScrollPane(this.console);
		
		this.controls = new PlaybackControls(this);
		this.config = new ParticleConfig(this);
		
		
		this.tabs = new JTabbedPane();
		this.tabs.addTab("Controls", null, this.controls, "Playback Controls");
		this.tabs.addTab("Particle Config", null, this.config, "Particle Configuration");
		
		
		//set up containers
		this.leftPane = new JSplitPane(JSplitPane.VERTICAL_SPLIT, this.slamvis, consoleScroll);
		this.mainSplit = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, this.leftPane, this.tabs);
		
		this.add(this.mainSplit);
		
		
		//Cleanup
		//Configure cleanup
		this.addWindowListener(new WindowAdapter() 
		{public void windowClosing(WindowEvent e) 
		{System.exit(0);}});
		
		//turn on frame
		this.setVisible(true);
		
			
	}


	public void actionPerformed(ActionEvent e) {
		if (e.getSource().equals(this.exit)){
			System.exit(0);
		} else if (e.getSource().equals(this.exportDatalog)){
			exportDatalog();
		} else if (e.getSource().equals(this.openDatalog)){
			openDatalog();
		}
		
	}

	
	private void openDatalog() {
		int ret = this.fc.showOpenDialog(this);
		
		ObjectInputStream in = null;
		if (ret == JFileChooser.APPROVE_OPTION){
			File file = this.fc.getSelectedFile();
			printToConsole("Opening Datalog: " + file.getAbsolutePath());

			try {
				in = new ObjectInputStream(
						new BufferedInputStream(
								new FileInputStream(file)));
				Object input = in.readObject();
				if (input instanceof Vector){
					this.controls.setData((Vector<Stim>)input);
					this.curLog = this.controls.getData();
					this.logName = file.getName();
					printToConsole("Datalog Opened: " + file.getName() + " with " + this.curLog.size() + " stims");
					this.setSubTitle(this.logName);
					this.controls.logReport();
				} else {
					printToConsole("Not a SLAM Datalog: " + file.getName());
				}
				in.close();
			} catch (FileNotFoundException e) {
				printToConsole("File: " + file.getAbsolutePath() + " not found!");
			} catch (IOException e) {
				printToConsole("Exception!");
				e.printStackTrace();
			} catch (ClassNotFoundException e) {
				printToConsole("Exception!");
				e.printStackTrace();
			}
		}
		
	}
	
	private void exportDatalog() {
		
		if (this.curLog.size() == 0){
			printToConsole("No Datalog or Empty Datalog Loaded!");
			return;
		}
		
		//create text
		printToConsole("Converting File...");
		StringBuilder str = new StringBuilder();
		str.append(this.logName + " Datalog - " + this.curLog.size() + "stims\n");
		for (int i = 0; i < this.curLog.size(); ++i){
			Stim cur = this.curLog.get(i);
			if (cur instanceof Control){
				str.append("CONTROL(" + i + ")\n");
				str.append(((Control)cur).toString() + "\n");
			} else if (cur instanceof PointMap){
				str.append("MEAUSUREMENT(" + i + ")\n");
				str.append(((PointMap)cur).toString() + "\n");
			} 
		}
		str.append("\nDatalog Complete");

		//save file
		JFileChooser fc = new JFileChooser();
		int ret = fc.showSaveDialog(this);
		if (ret == JFileChooser.APPROVE_OPTION){
			File file = fc.getSelectedFile();
			OutputStreamWriter out = null;
			
			try {
				FileOutputStream fw = new FileOutputStream(file);
				out = new OutputStreamWriter(fw);
				out.write(str.toString());
				fw.close();
				printToConsole("Export Successful!");
			} catch (FileNotFoundException e) {
				printToConsole("File Not Found: " + file.getAbsolutePath());
				e.printStackTrace();
			} catch (IOException e) {
				printToConsole("Exception!");
				e.printStackTrace();
			}
		}
	}


	/**
	 * Writes to the console by appending
	 * @param s
	 */
	public void printToConsole(String s) {
		this.console.append("\n" + s);
	}
	
	/**
	 * Sets the string following the main title
	 * @param s
	 */
	private void setSubTitle(String s){
		this.setTitle("SLAM Offline Analysis - " + s);
	}


	public SLAMVisPanel getSlamvis() {
		return slamvis;
	}
	
	public Particle getParticle(){
		return this.particle;
	}
	
	public void updateParticle(Particle p){
		this.particle = p;
		this.slamvis.paint(this.slamvis.getGraphics());
	}


	public Vector<Stim> getCurLog() {
		return curLog;
	}


	public void setCurLog(Vector<Stim> curLog) {
		this.curLog = curLog;
	}
	
	
	

}
