package edu.wpi.ari.glados.vision;

import java.awt.Dimension;
import java.awt.Point;
import java.awt.image.BufferedImage;
import java.util.Vector;

import javax.vecmath.Vector3d;

import edu.wpi.ari.glados.vision.vidar.Detector;
import edu.wpi.ari.glados.vision.vidar.Matcher;

/**
 * A manager for managing the vision system. 
 * Handles the main loop of the vision system, calling the
 * detector and matcher functions as necessary, as well as implementing 
 * the IVisionObervable interface. 
 * @Note The Manager also contains the global defines MATCH_THRESHOLD and MATCH_RELUCTANT
 * ,which modify how the matcher works under different conditions. 
 * @author robert breznak, Zack A. Kleinfeld, Alex Cunningham
 */
public class VisionManager extends Thread implements IVisionObervable {
	private Vector<VisionObserver> listeners;
	
	private ImageCanvas leftImage;
	private ImageCanvas rightImage;
	
	private boolean run = true;
	private boolean connected;
	/**MATCH_THRESHHOLD
	 * Determines how good matches must be to count as the same point. 
	 * @Note Good values must take into consideration how many pixels are in an image.
	 * If values are too large or two small, large disparities will not be found. 
	 * 
	 */
	public int MATCH_THRESHHOLD = 10; //How good should matches be? 
	//Good values are between width/64 and width/128
	/**MATCH_RELUCTANT
	 * Determines if a point is to be removed after use, to not be used more than once.
	 * possible use is to eliminate concentration on one strong feature. 
	 */
	public boolean MATCH_RELUCTANT = false; //Should be remove points once we use them?
	
	/**
	 * Constructs a new vision manager.
	 * Decides at this point what resolution will be taken from the camera.
	 * 
	 * @param host the host to connect to
	 * @param leftPort the port to communicate across for the left camera
	 * @param rightPort the port to communicate across for the right camera
	 * @param fps the maximum fps rate to capture images at
	 * @param d the size of the windows to display
	 */
	public VisionManager(String host, int leftPort, int rightPort, int fps, Dimension d) {
		leftImage = new ImageCanvas(host, leftPort, fps, d.height, d.width);
		rightImage = new ImageCanvas(host, rightPort, fps, d.height, d.width);
		
		listeners = new Vector<VisionObserver>();
	}
	
	/**
	 * Change the configuration of the camera connections
	 * Can change resolution at this point, but changes must propagate to both cameras
	 * at the same time. 
	 * @note This function makes sure the resolutions between the two cameras are the same.
	 * 
	 * @param host the new host to connect to
	 * @param leftPort the port to communicate across for the left camera
	 * @param rightPort the port to communicate across for the right camera
	 * @param fps the maximum fps rate to capture images at
	 * @param d the size of the windows to display
	 */
	public void configure(String host, int leftPort, int rightPort, int fps, Dimension d) {
			// connect the cameras.
			leftImage = new ImageCanvas(host, leftPort, fps, d.height, d.width);
			rightImage = new ImageCanvas(host, rightPort, fps, d.height, d.width);
	}
	
	/**
	 * Connect the cameras and start calculations on them
	 * Kicks off the main vision thread.
	 * Calls connect on each ImageCanvas.
	 */
	public void connect() {
		leftImage.connect();
		rightImage.connect();
		
		start();
		
		connected = true;
	}
	
	/**
	 * Disconnect the cameras and stop calculations.
	 * Calls disconnect on each ImageCanvas and sets halt flags.
	 */
	public void disconnect() {
		run = false;
		leftImage.disconnect();
		rightImage.disconnect();
		connected = false;
	}
	
	/**
	 * Get the left camera feed
	 * @return the left camera feed
	 */
	public ImageCanvas getLeft() {
		return leftImage;
	}
	
	/**
	 * Get the right camera feed 
	 * @return the right camera
	 */
	public ImageCanvas getRight() {
		return rightImage;
	}
	
	/** Fires all calculated feature points, updating all observers with the new value.
	 * @param points A Vector of Vector3d with all calculated features.
	 */
	private void fireUpdate(Vector<Vector3d> points) {
		for(VisionObserver o : listeners) {
			o.VisionReceivedValues(points);
		}
	}
	
	/**
	 * Called by the new thread. 
	 * 
	 * Calculates the new disparity values and then fires the results to all of
	 * the listening observers.
	 */
	public void run() {
		while(run) {
			// give some time for an interruption and for the images to update
			// if interrupted, disconnect everything and stop all calculations
			try {
				// don't do anything if the cameras aren't connected or the thread is
				// already running
				// if(run || !connected) { return; }
				
				run = true;
				
				Vector<Vector3d> feats = matchSeq();
				
				for(Vector3d x : feats) {
					System.err.println(x);
				}
				
				fireUpdate(feats);
				
				sleep(500);
			} catch (InterruptedException e) {
				disconnect();
				return;
			}
		}
	}

	/**The actual image manipulation occurs here.
	 * The sequence is this:
	 * 1. A Matcher and Detector are created.
	 * 2. Both images are grabbed.
	 * 3. Feature Detection applied to both images.
	 * 4. Feature lists are matched.
	 * 
	 * @Note Any Feature Detection algorithm replacment will have to replace this function.
	 * @return Calculated Features
	 */
	private Vector<Vector3d> matchSeq() {
		Detector d = new Detector();
		Matcher m = new Matcher();
		BufferedImage start;
		do{
			start = getLeft().getBufferedImage();
		}
		while(start==null);
		
		BufferedImage start2 = getRight().getBufferedImage();
		
		BufferedImage dest = new BufferedImage(start.getWidth(), start.getHeight(), start.getType());
		d.edgeDetect(start, dest);
		BufferedImage dest2 = new BufferedImage(start.getWidth(), start.getHeight(), start.getType());
		d.lineDetect(dest, dest2);
		BufferedImage dest3 = new BufferedImage(start.getWidth(), start.getHeight(), start.getType());
		d.highDetect(dest2, dest3);
		
		
		BufferedImage dest02 = new BufferedImage(start.getWidth(), start.getHeight(), start.getType());
		d.edgeDetect(start2, dest02);
		BufferedImage dest22 = new BufferedImage(start.getWidth(), start.getHeight(), start.getType());
		d.lineDetect(dest02, dest22);
		BufferedImage dest32 = new BufferedImage(start.getWidth(), start.getHeight(), start.getType());
		d.highDetect(dest22, dest32);
		
		Vector<Point> result0 = d.featureDetect(dest3, 1, 100);
		Vector<Point> result1 = d.featureDetect(dest32, 1, 100);
		
		Vector<Vector3d> feats = m.match(result0, result1, MATCH_THRESHHOLD, 140, 1.0, 1.0, MATCH_RELUCTANT);
		return feats;
	}
	
	public void addListener(VisionObserver l) {
		listeners.add(l);
	}

	public void removeAll() {
		listeners.clear();
	}

	public void removeListener(VisionObserver l) {
		listeners.remove(l);
	}
	
}
