package edu.wpi.ari.glados.vision.vidar;
import java.awt.Point;
import java.awt.image.BufferedImage;
import java.util.Vector;

import javax.vecmath.Vector3d;

/** Matcher is responsible for calculating the likelyhood two features are the same,
 * then must calculate their distance from the observer.
 * @author Zack A. Kleinfeld
 *
 */
public class Matcher {
	
	/**Maximum distance two points can be found at. A replacment for infinity.
	 * 
	 */
	public double MAX_DISTANCE = 200000.0;
	
	/**Camera field of view, in degrees.
	 * 
	 */
	public double FOV = 42.0;
	
	/**Pixel offset on the x axis relative to the right camera.
	 * The image will be shifted right this number of degrees.
	 * 
	 */
	public int LEFT_CAM_OFFSET_X = 150;
	
	/**Pixel offset on the y axis relative to the right camera.
	 * The image will be shifted down this number of degrees.
	 * 
	 */
	public int LEFT_CAM_OFFSET_Y = 50;
	
	
	/**The main feature matching function.
	 * 
	 * @param lhs Vector of LHS features.
	 * @param rhs Vector of RHS features.
	 * @param threshhold How close feature must be to be matched.
	 * @param base Camera spread, in mm.
	 * @param focalLength focal length of both cameras, in mm
	 * @param converstionFactor factor to convert pixels to mm
	 * @param destructive Should we remove a feature when matched?
	 * @return
	 */
	public Vector<Vector3d> match(Vector<Point> lhs, 
			Vector<Point> rhs,
			double threshhold,
			int base,
			double focalLength,
			double converstionFactor,
			boolean destructive){
		Vector<Vector3d> results = new Vector<Vector3d>();
		for (Point p1 : lhs){
			for (Point p2 : rhs){
				if ( Math.sqrt(((p1.x + LEFT_CAM_OFFSET_X) - p2.x)^2 + 
						((p1.y + LEFT_CAM_OFFSET_Y) -p2.y)^2) <= threshhold){
					results.add(new Vector3d(p1.x-90, p1.y, 
							calcDistance((p1.x -p2.x), base,
									focalLength, converstionFactor)));
					if (destructive){
						rhs.remove(p1);
					}
					break;
				}
			}
		}
		return results;
	}
	
	/**Calculate the distance using the disparity. Uses the standard method.
	 * 
	 * @param disperity The pixel-wise disparity between the two images.
	 * @param base Camera spread, in mm.
	 * @param focalLength focal length of both cameras, in mm
	 * @param converstionFactor factor to convert pixels to mm
	 * @return Distance in arbitrary units.
	 * 
	 * @Note In our system, the units this returns in are a factor largely of the focal length.
	 * By dimensional analysis, the result can be determined if all units are known.
	 */
	public double calcDistance(int disperity, int base,
			double focalLength, double converstionFactor){
		if (disperity == 0.0){
			return MAX_DISTANCE; 
		}else{
		return Math.abs((base*focalLength)/disperity*converstionFactor);
		}
	}
	
	/** Using the FOV defined, calculate the real vector to a point, adjusted 
	 * from X, Y, Distance.
	 * @param vec X, Y, and Z values calculated from calcDistance
	 * @param img image (used for its dimensions) from a camera.
	 * @return Real vector to the point.
	 */
	public Vector3d createVector(Vector3d vec, BufferedImage img){
		//the Cameras we're using span 42 degrees FOV.
		double scale = FOV/img.getWidth()/2;
		double newX = (vec.x - img.getWidth()/2)*scale;
		double newY = (vec.y - img.getHeight()/2)*scale;
		return new Vector3d(newX, newY, vec.z);
	}
}
